import java.applet.Applet;
import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import java.awt.geom.*;
import java.math.*;

/**This is our class for members of Q[phi][psi]. 
   We are only going to consider dyadic rationals
   in this field.  That is, the denominator is
   always a power of 2.
   Here psi = (1+ Sqrt[13])/2*/

public class QI2 {
    QI a;
    QI b;   
    long e; //exponent:  // (a + b psi)/2^e
    boolean EASY;       // when SPECIAL=true, we have b=0 and e=0.

    public QI2() {}


    /**These are the construction methods*/
    public QI2(long A,long B) {
	this.a=new QI(A,B);
	this.b=new QI(0,0);
	this.e=0;
	this.EASY=true;
    }


    public QI2(QI A,QI B,long E) {
	this.a=new QI(A);
	this.b=new QI(B);
	this.e=E;
	EASY=false;
    }

    public QI2(QI2 A) {
	this.a=new QI(A.a);
	this.b=new QI(A.b);
	this.e=A.e;
	EASY=A.EASY;
    }

    public static QI2 zero() {
	QI2 A=new QI2(new QI(0,0),new QI(0,0),0);
	A.EASY=true;
	return(A);
    }

    public static QI2 one() {
	QI2 A=new QI2(new QI(1,0),new QI(0,0),0);
	A.EASY=true;
	return(A);
    }

    public static BigInteger exp2(long k) {
	BigInteger k1=new BigInteger("1");
	BigInteger k2=new BigInteger("2");
	for(int i=0;i<k;++i) k1=k1.multiply(k2);
	return(k1);
    }

    public QI2(QI A) {
	this.a=new QI(A);
	this.b=new QI(0,0);
	this.e=0;
	this.EASY=true;
    }

    public static QI2 plus(QI2 x,QI2 y) {	

	if((x.EASY==true)&&(y.EASY==true)) {
	  QI2 z=zero();
          z.a=QI.plus(x.a,y.a);
	  z.EASY=true;
	  return(z);
	}

        QI2 z=new QI2();
	if(x.e<y.e) {
	    BigInteger q=exp2(y.e-x.e);
	    QI xa=QI.times(x.a,new QI(q));
	    QI xb=QI.times(x.b,new QI(q)); 
            z.a=QI.plus(xa,y.a);
	    z.b=QI.plus(xb,y.b);
	    z.e=y.e;
	    return(z);
	}  
	if(x.e>y.e) {
	    BigInteger q=exp2(x.e-y.e);
	    QI ya=QI.times(y.a,new QI(q));
	    QI yb=QI.times(y.b,new QI(q)); 
            z.a=QI.plus(x.a,ya);
	    z.b=QI.plus(x.b,yb);
	    z.e=x.e;
	    return(z);
	}  

        z.a=QI.plus(x.a,y.a);
	z.b=QI.plus(x.b,y.b);
	z.e=x.e;
	return(z);
    }


    public QI2 negate() {
	QI2 V=new QI2(a.negate(),b.negate(),e);
	V.EASY=EASY;
	return(V);
    }


    public static QI2 minus(QI2 x,QI2 y) {
	return(plus(x,y.negate()));
    }


    public static QI2 times(QI2 x,QI2 y) {  

	if((x.EASY==true)&&(y.EASY==true)) {
	  QI2 z=zero();
          z.a=QI.times(x.a,y.a);
	  z.EASY=true;
	  return(z);
	}

        QI2 z=new QI2();
	QI three=new QI(3,0);
	z.a=QI.times(x.a,y.a);
	z.a=QI.plus(z.a,QI.times(three,QI.times(x.b,y.b)));
	z.b=QI.times(x.a,y.b);
	z.b=QI.plus(z.b,QI.times(x.b,y.a));
	z.b=QI.plus(z.b,QI.times(x.b,y.b));
	z.e=x.e+y.e;
	return(z);
    }

    public static QI2 pow(QI2 x,int j) {
	QI2 y=new QI2(new QI(1,0),new QI(0,0),0);
	for(int i=0;i<j;++i) {
	    y=times(y,x);
	}
	y.EASY=x.EASY;
	return(y);
    }

    /**vector arithmetic*/

    public static QI2[] plus(QI2[] X,QI2[] Y) {
	QI2[] Z={plus(X[0],Y[0]),plus(X[1],Y[1])};
	return(Z);
    }

    public static QI2[] minus(QI2[] X,QI2[] Y) {
	QI2[] Z={minus(X[0],Y[0]),minus(X[1],Y[1])};
	return(Z);
    }


    public static QI2 dot(QI2[] X,QI2[] Y) {
	QI2 Z=plus(times(X[0],Y[0]),times(X[1],Y[1]));
	return(Z);
    }

    /**The sign*/

    public boolean isZero() {
	if(EASY==true) return(a.isZero());
	if(a.isZero()==false) return(false);
	if(b.isZero()==false) return(false);
	return(true);
    }



    public boolean isPositiveOrZero() {
	if(isZero()==true) return(true);
	if(isPositive()==true) return(true);
	return(false);
    }

    public int sign() {
	if(isZero()==true) return(0);
	boolean test1=isPositive();
	if(test1==true) return(1);
	QI2 X=this.negate();
	boolean test2=X.isPositive();
	if(test2==true) return(-1);
	return(99);
    }


    /**We first produce integers s,t,u,v, so that the
       number is (s+t sq(5)) + (u + v sq(5)) sq(13).
       Then we analyze the situation as in the
       quadratic case.*/

    public boolean isPositive() {
	if(EASY==true) return(a.isPositive());
	BigInteger ZERO=new BigInteger("0");
	BigInteger TWO=new BigInteger("2");
	BigInteger FOUR=new BigInteger("4");
	BigInteger aa=this.a.a;
	BigInteger ab=this.a.b;
	BigInteger ba=this.b.a;
	BigInteger bb=this.b.b;
	BigInteger s=FOUR.multiply(aa);
	s=s.add(TWO.multiply(ab));
	s=s.add(TWO.multiply(ba));
	s=s.add(bb);
	BigInteger t=TWO.multiply(ab);
	t=t.add(bb);
	BigInteger u=TWO.multiply(ba);
	u=u.add(bb);
	BigInteger v=bb;
	return(QISign.isPositive(s,t,u,v));
    }

    public double toDouble() {
	if(EASY==true) return(a.toDouble());
	double aa=a.toDouble();
	double bb=b.toDouble();
	double psi=(1+Math.sqrt(13))/2;
	return((aa+bb*psi)/Math.pow(2,e));
    }


    public static QI2 recognize(double t) {
	QI2 q=new QI2();
	int i0=0;
	int j0=0;
	int k0=0;
	int l0=0;
	double min=1000;
	for(int i=-15;i<16;++i) {
	    for(int j=-15;j<16;++j) {
		for(int k=-15;k<16;++k) {
		    for(int l=-15;l<16;++l) {
			QI X=new QI(i,j);
			QI Y=new QI(k,l);
			q=new QI2(X,Y,1);
			double u=q.toDouble();
			double test=Math.abs(t-u);
			if(test<min) {
			    i0=i;j0=j;k0=k;l0=l;
			    min=test;
			}
		    }
		}
	    }
	}
	QI X=new QI(i0,j0);
	QI Y=new QI(k0,l0);
        q=new QI2(X,Y,1);
	System.out.println("min "+min);
	return(q);
    }

    public String toStringLaTeX() {
	if(EASY==true) {
	    String S="";
	    S=S+a.a.toString()+","+a.b.toString();
	    return(S);
	}
	return("special");
    }

    /**Our computer program only has 2 points in the
       partition which do not lie in the quadratic
       number field Q[phi].  We just do something
       special for these points.*/

    public String toString() {
	if(EASY==true) return(a.toString());
	double t=toDouble();
	if(Math.abs(t-6.09)<.1) return("b(\u03A8)");
	if(Math.abs(t+.429)<.1) return("b(1-\u03A8)");
	if(Math.abs(t-1.125)<.1) return("-\u03D5^2 b(1-\u03A8)");
	return("Q[\u03D5,\u03A8]");
    }




    public void print() {
	if(EASY==true) {
	    a.print();
	    return;
	}
	System.out.print("(");
	a.print();
	System.out.print(") + (");
	b.print();
	System.out.print(") psi /2^");
	System.out.print(e+"  ");
    }

    public void println() {
	print();
	System.out.println("");
    }
}



