import java.applet.Applet;
import java.awt.event.*;
import java.awt.*;


public class QI2Differential {


    /**for ease of notation we copy some routines from other classes*/
    public static QI2RationalFunction times(QI2RationalFunction A,QI2RationalFunction B) {
	return(QI2RationalFunction.times(A,B));
    }

    public static QI2RationalFunction  plus(QI2RationalFunction A,QI2RationalFunction B) {
	return(QI2RationalFunction.plus(A,B));
    }

    public static QI2RationalFunction  minus(QI2RationalFunction A,QI2RationalFunction B) {
	return(QI2RationalFunction.minus(A,B));
    }

    public static QI2Polynomial times(QI2Polynomial A,QI2Polynomial B) {
	return(QI2Polynomial.times(A,B));
    }

    public static QI2Polynomial  plus(QI2Polynomial A,QI2Polynomial B) {
	return(QI2Polynomial.plus(A,B));
    }

    public static QI2Polynomial  minus(QI2Polynomial A,QI2Polynomial B) {
	return(QI2Polynomial.minus(A,B));
    }





    /**partial derivatives of a polynomial*/

    public static QI2Polynomial d0(QI2Polynomial P) {
	QI2Polynomial Q=new QI2Polynomial();
	int[] b=P.getBound();
	for(int i=0;i<b[0];++i) {
	    for(int j=0;j<b[1];++j) {
		Q.c[i][j]=QI2.times(P.c[i+1][j],new QI2(i+1,0));
	    }
	}
	return(Q);
    }

    public static QI2Polynomial d1(QI2Polynomial P) {
	QI2Polynomial Q=new QI2Polynomial();
	int[] b=P.getBound();
	for(int i=0;i<b[0];++i) {
	    for(int j=0;j<b[1];++j) {
		Q.c[i][j]=QI2.times(P.c[i][j+1],new QI2(j+1,0));
	    }
	}
	return(Q);
    }

    public static QI2Polynomial d(int i,QI2Polynomial P) {
	if(i==0) return(d0(P));
	if(i==1) return(d1(P));
	return(null);
    }

    /**partial derivatives of a rational function*/

    public static QI2RationalFunction d(int k,QI2RationalFunction F) {
	QI2Polynomial dP=d(k,F.P);
	QI2Polynomial dQ=d(k,F.Q);
	QI2RationalFunction G=new QI2RationalFunction();
	G.P=minus(times(dP,F.Q),times(dQ,F.P));
	G.Q=times(F.Q,F.Q);
	return(G);
    }

    /**matrix of first partials for a rational map*/

    public static QI2RationalFunction[][] D(QI2RationalMap X) {
	QI2RationalFunction[][] F=new QI2RationalFunction[2][2];
	for(int i=0;i<2;++i) {
	    for(int j=0;j<2;++j) {
		F[i][j]=d(j,X.F[i]);
	    }
	}
	return(F);
    }


    /**Jacobian*/

    public static QI2RationalFunction detD(QI2RationalMap X) {
	QI2RationalFunction[][] DX=D(X);
	QI2RationalFunction F=new QI2RationalFunction();
	F.P=minus(times(DX[0][0].P,DX[1][1].P),times(DX[0][1].P,DX[1][0].P));
	F.Q=times(DX[0][0].Q,DX[1][0].Q);
	return(F);
    }

    /**determinant*/

    public static QI2RationalFunction det(QI2RationalFunction[][] m) {
	QI2RationalFunction D=new QI2RationalFunction();
	D=minus(times(m[0][0],m[1][1]),times(m[1][0],m[0][1]));
	return(D);
    }

    /**square norm bounds: sum of squares of first ROW.
       The denominators DX[0][0].Q and DX[0][1].Q are the same,
       so the denominator of our rational function can be 
       computed just using DX[0][0].Q.  This is why we are
       computing the rows separately, rather than together.
       We are trying to cut down on the size of the polynomial*/

    public static QI2RationalFunction sumOfSquares1D(QI2RationalMap X) {
	QI2RationalFunction[][] DX=D(X);
	QI2RationalFunction F=new QI2RationalFunction();
	F.P=plus(times(DX[0][0].P,DX[0][0].P),times(DX[0][1].P,DX[0][1].P));
	F.Q=times(DX[0][0].Q,DX[0][0].Q);
	return(F);
    }

    /**square norm bounds: sum of squares of second ROW.
       Same remarks apply to the calculation of the denominator.*/

    public static QI2RationalFunction sumOfSquares2D(QI2RationalMap X) {
	QI2RationalFunction[][] DX=D(X);
	QI2RationalFunction F=new QI2RationalFunction();
	F.P=plus(times(DX[1][0].P,DX[1][0].P),times(DX[1][1].P,DX[1][1].P));
	F.Q=times(DX[1][0].Q,DX[1][0].Q);
	return(F);
    }



    /**This computes DF(a0,a1)*/
    public static QI2RationalFunction[] DIFF(QI2RationalMap X,QI2 a0,QI2 a1) {
	QI2RationalFunction[][] DX=D(X);
	QI2RationalFunction[] B=new QI2RationalFunction[2];
	QI2Polynomial A0=new QI2Polynomial(a0);
	QI2Polynomial A1=new QI2Polynomial(a1);
	for(int i=0;i<2;++i) {
	    B[i]=new QI2RationalFunction();
	    B[i].P=plus(times(A0,DX[i][0].P),times(A1,DX[i][1].P));	
            B[i].Q=DX[i][0].Q;
	}
	return(B);
    }


}



